#!/usr/bin/env python3

###########################################################################################
#  package:   pNbody
#  file:      libvazdekis.py
#  brief:     Vazdekis luminosities
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
###########################################################################################


import numpy as np

from .libSSPluminosity import SSPLuminosities


#####################################################
def read_ascii(
        file,
        columns=None,
        lines=None,
        dtype=float,
        skipheader=False,
        cchar='#'):
    #####################################################
    """[X,Y,Z] = READ('FILE',[1,4,13],lines=[10,1000])
    Read columns 1,4 and 13 from 'FILE'  from line 10 to 1000
    into array X,Y and Z

    file is either fd or name file

    """

    def RemoveComments(l):
        if l[0] == cchar:
            return None
        else:
            return l

    def toNumList(l):
        return list(map(dtype, l))

    if isinstance(file, str):
        f = open(file, 'r')
    else:
        f = file

    # read header while there is one
    while True:
        fpos = f.tell()
        header = f.readline()
        if header[0] != cchar:
            f.seek(fpos)
            header = None
            break
        else:
            if skipheader:
                header = None
            else:
                # create dict from header
                header = str.strip(header[2:])
                elts = str.split(header)
                break

    # now, read the file content
    lines = f.readlines()

    # remove trailing
    lines = list(map(str.strip, lines))

    # remove comments
    #lines = map(RemoveComments, lines)

    # split
    lines = list(map(str.split, lines))

    # convert into float
    lines = list(map(toNumList, lines))

    # convert into array
    lines = np.array(list(map(np.array, lines)))

    # np.transpose
    lines = np.transpose(lines)

    if header is not None:
        iobs = {}
        i = 0
        for elt in elts:
            iobs[elt] = i
            i = i + 1

        vals = {}
        for key in list(iobs.keys()):
            vals[key] = lines[iobs[key]]

        return vals

    # return
    if columns is None:
        return lines
    else:
        return lines.take(axis=0, indices=columns)


BAND_DIC = {"U":18,"B":19,"V":20,"R":21,"I":22}

class VazdekisLuminosities(SSPLuminosities):
  
    
    def Read(self):
        """
        read file and create a data table
        """
        self.data = read_ascii(self.file, cchar='#', skipheader=True)

    def CreateMatrix(self):
        """
        from data extract
        metalicites (zs)
        ages (ages)
        and ML (vs)
        """
        zs = self.data[1, :]
        ages = self.data[2, :]
        mls = self.data[BAND_DIC[self.band], :]

        # create borders
        Zs = [-2.3152, -1.7129, -1.3146, -0.7052, -0.3960, 0.0000, 0.2223]
        Ages = np.compress(Zs[1] == zs, ages)

        MatLv = np.zeros((len(Zs), len(Ages)))

        for iZ, Z in enumerate(Zs):

            zs, ages, mls

            c = (zs == Z)

            age = np.compress(c, ages)
            ml = np.compress(c, mls)
            Lv = 1 / ml

            if len(age) < len(Ages):
                n = len(Ages) - len(age)
                age = np.concatenate((Ages[:n], age))
                Lv = np.concatenate((1e-10 * np.ones(n), Lv))

            MatLv[iZ, :] = Lv

        self.MatLv = MatLv
        self.Ages = Ages
        self.Zs = Zs



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      h5utils.py
#  brief:     Defines usefull function related to the hdf5 format
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from . import mpi
import h5py
import numpy as np

try:
    from mpi4py import MPI
except BaseException:
    MPI = None

def openFile(fname,mode='r'):
    """
    open an hdf5 file
    """
    if mpi.mpi_NTask() > 1:
        fd = h5py.File(fname,mode,driver="mpio",comm=MPI.COMM_WORLD)
    else:
        fd = h5py.File(fname,mode)

    return fd

def closeFile(fd):
    """
    close an hdf5 file
    """
    mpi.mpi_barrier()
    fd.close()


def get_particles_limits(size):
    """ Gives the limits for a thread.
    In order to get the particles, slice them like this pos[start:end].
    :param int i: Particle type
    :returns: (start, end)
    """
    nber = float(size) / mpi.mpi_NTask()
    start = int(mpi.mpi_ThisTask() * nber)
    end = int((mpi.mpi_ThisTask() + 1) * nber)
    return start, end    


def get_particles_limits_from_npart(npart_i):
    """ Gives the limits for a thread.
    In order to get the particles, slice them like this pos[start:end].
    :param npart_tot_i: total particle of type i
    :returns: (start, end)
    """
    nber = float(npart_i) / mpi.mpi_NTask()
    start = int(mpi.mpi_ThisTask() * nber)
    end = int((mpi.mpi_ThisTask() + 1) * nber)
    return start, end
        


def get_npart_from_dataset(fname,key,ntype,ptypes=None):
  '''
  Get an array containing the number of particles for each particle type
  from the block named key.
  
  fname  : the file name
  key    : the block to read
  ntype  : number of particle types 
  '''

  if ptypes is None:
    ptypes = list(range(ntype))
    
  h5_key = key 
  
  # open the file
  fd = openFile(fname)    
  
  npart = np.zeros(ntype,np.uint32)
  
  # loop over particle types
  for i_type in ptypes:
    PartType = "PartType%d" % i_type
    
    if PartType in fd:
      block = fd[PartType]
      
      # if the dataset is present
      if h5_key in fd[PartType].keys():
        
        data_lenght   = block[h5_key].len()
      
        idx0=0
        idx1=data_lenght
        if mpi.mpi_NTask() > 1:
          idx0, idx1 = get_particles_limits(data_lenght)
        
        npart[i_type] = idx1-idx0   
        
      else:
        raise error.FormatError("%s is not in the file."%(h5_key))     
  
  
  # close the file
  closeFile(fd)
  
  return npart



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      plummer.py
#  brief:     Defines plummer profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
plummer model
"""

import numpy as np


def Potential(M, a, r, G=1.):
    """
    Plummer Potential
    """
    return -G * M / np.sqrt(r**2 + a**2)


def dPotential(M, a, r, G=1.):
    """
    Plummer first derivative of Potential
    """
    return G * M * r * (r**2 + a**2)**(-3. / 2.)


def CumulativeMass(M, a, r, G=1.):
    """
    Plummer cumulative mass
    """
    return M * r**3 / (r**2 + a**2 )**(3./2.)


def TotalMass(M, a, G=1.):
    """
    Plummer total mass
    """
    return M 
    


def Vcirc(M, a, r, G=1.):
    """
    Plummer circular velocity
    """
    return np.sqrt(G * M * r**2 * (r**2 + a**2)**(-3. / 2.))


def Density(M, a, r, G=1.):
    """
    Plummer Density
    """
    return (3. * M / (4. * np.pi * a**3)) * (1 + (r / a)**2)**(-5. / 2.)


def LDensity(M, a, r, G=1.):
    """
    Plummer Linear Density
    """
    return (4 * np.pi * r**2) * (3. * M / (4. * np.pi * a**3)) * \
        (1 + (r / a)**2)**(-5. / 2.)


def Sigma(M, a, r, G=1.):
    """
    Return sigma (radial) from Jeans equation : 1/rho Int( rho * drPhi * dr )
    """
    sigma = 1. / (8 * np.pi * Density(G, M, a, r)) * \
        G * M**2 * a**2 / (r**2 + a**2)**3
    return np.sqrt(sigma)



###########################################################################################
#  package:   pNbody
#  file:      thermodyn.py
#  brief:     thermodynamics functions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from astropy import constants as c
from astropy import units as u
#from pNbody.units import unitSystem
from pNbody import thermodynlib


# default adiabatic index
gamma = 5/3.


'''
lu = unitSystem()

def ToEnergySpec(x):
  if type(x) is not u.quantity.Quantity:
    x = x*lu.UnitEnergySpec
  return x

def ToDensity(x):
  if type(x) is not u.quantity.Quantity:
    x = x*lu.UnitDensity
  return x

def ToTemperature(x):
  if type(x) is not u.quantity.Quantity:
    x = x*lu.UnitTemperature
  return x  
'''

def checkQuantity(x):
  if type(x) is not u.quantity.Quantity:
    raise("must be an astropy Quantity")



# tabulation of the mean weight (from Grackle)
nt = 10
tt = np.array([1.0e+01, 1.0e+02, 1.0e+03, 1.0e+04, 1.3e+04,2.1e+04, 3.4e+04, 6.3e+04, 1.0e+05, 1.0e+09])
mt = np.array([1.18701555,1.15484424,1.09603514,0.9981496,0.96346395,0.65175895,0.6142901,0.6056833,0.5897776,0.58822635])



def MeanMolecularWeight(T):
  """
  mean molecular weight as a function of the Temperature
  T: temperature
  """
  return thermodynlib.MeanWeightT(T)


def UN_T(T):
  """
  UN_T(T) = energy normalized as a function of the Temperature
        = T/mu(T)
  T: temperature
  """
  return T / MeanMolecularWeight(T)


# tabulation of the normalized energy vs T
unr = UN_T(tt)


def T_UN(UN):
  """
  T_UN(UN) = temperature vs energy normalized
  
  inverse of UN_T(U)
  
  """
  
  if isinstance(UN, np.ndarray):

      T = np.zeros(len(UN))
      for i in range(len(UN)):
          T[i] = Tun(UN[i])

  else:

      logu = np.log(UN)
      uuu = np.exp(logu)

      if uuu < unr[0]:
          j = 1
      else:
          for j in range(1, nt):
              if (uuu > unr[j - 1]) and (uuu <= unr[j]):
                  break

      slope = np.log(tt[j] / tt[j - 1]) / np.log(unr[j] / unr[j - 1])
      T = np.exp(slope * (logu - np.log(unr[j])) + np.log(tt[j]))

  return T
  


def Temperature_fromEnergySpec(energy):
  """
  temperature from energy
  energy  : specific energy
  """
  # add units if needed
  #energy = ToEnergySpec(energy)
  checkQuantity(energy)

  # normalized energy : UN is in K
  UN = (gamma - 1) * c.m_p/c.k_B * energy 
    
  # compute temperature from normalized energy
  T = thermodynlib.Tun(UN.to(u.K).value)*u.K
  
  return T


def Temperature_fromSoundSpeed(density, sound_speed):
  """
  Temperature from Sound Speed
  density : density
  sound_speed  : sound speed  
  """
  energy = EnergySpec_fromSoundSpeed(density, sound_speed)
  return Temperature_fromEnergySpec(energy)


def Temperature_fromJeansMass(density, jeans_mass):
  """
  Temperature from Jeans Mass
  density : density
  jeans_mass  : jeans mass
  """
  energy = EnergySpec_fromJeansMass(density,jeans_mass)
  return Temperature_fromEnergySpec(energy)  







def Energy_fromTemperature(temperature):
  """
  temperature from energy
  energy  : specific energy
  """
  # add units if needed
  #temperature = ToTemperature(temperature)
  checkQuantity(temperature)
  
  U = 1/(gamma - 1.) * (c.k_B/c.m_p) * UN_T(temperature.to(u.K).value)*u.K

  #return ToEnergySpec(U)
  return U


def EnergySpec_fromSoundSpeed(density, sound_speed):
  """
  Energy Spec from Sound Speed
  density : density
  sound_speed  : sound speed  
  """
  checkQuantity(density)
  checkQuantity(sound_speed)  

  return sound_speed**2/((gamma - 1.) * gamma)


def EnergySpec_fromJeansMass(density, jeans_mass):
  """
  Energy Spec from Jeans Mass
  density : density
  jeans_mass  : jeans mass
  """
  Cs = SoundSpeed_fromJeansMass(density,jeans_mass)
  return EnergySpec_fromSoundSpeed(density,Cs)
  
  
  




def Pressure_fromEnergySpec(density, energy):
  """
  pressure from density and energy
  density : density
  energy  : specific energy
  """
  # add units if needed
  #density = ToDensity(density)
  #energy  = ToEnergySpec(energy)
  checkQuantity(density)
  checkQuantity(energy)

  return (gamma-1) * energy * density
    
    
def Pressure_fromTemperature(density, temperature):
  """
  pressure from density and temperature
  density : density
  temperature  : temperature
  """
  # add units if needed
  #density = ToDensity(density)
  #temperature  = ToTemperature(temperature)
  checkQuantity(density)
  checkQuantity(temperature)
    
  energy = Energy_fromTemperature(temperature)

  return Pressure_fromEnergySpec(density, energy)






def SoundSpeed_fromEnergySpec(density, energy):
  """
  Sound Speed
  density : density
  energy  : specific energy  
  """
  # add units if needed
  #density = ToDensity(density)
  #energy  = ToEnergySpec(energy)
  checkQuantity(density)
  checkQuantity(energy)  
  
  return np.sqrt((gamma - 1.) * gamma * energy)


def SoundSpeed_fromTemperature(density, temperature):
  """
  Sound Speed
  density : density
  energy  : specific energy  
  """
  energy = Energy_fromTemperature(temperature)  
  return SoundSpeed_fromEnergySpec(density, energy)
  

def SoundSpeed_fromJeansMass(density, jeans_mass):
  """
  Sound Speed
  density : density
  jeans_mass  : jeans mass 
  """
  # add units if needed
  #density = ToDensity(density)
  #energy  = ToEnergySpec(energy)
  checkQuantity(density)
  checkQuantity(jeans_mass)
  
  return (jeans_mass * (6 * c.G**(3 / 2.) * np.sqrt(density)) / np.pi**(5 / 2.))**(1/3.)






     
def JeansLength_fromEnergySpec(density, energy):
  """
  Jeans Length
  density : density
  energy  : specific energy  
  """  
  Cs = SoundSpeed_fromEnergySpec(density,energy)
  return Cs * np.sqrt(np.pi / (c.G * density))  


def JeansLength_fromTemperature(density, temperature):
  """
  Jeans Length
  density : density
  temperature  : temperature 
  """
  energy = Energy_fromTemperature(temperature)  
  return JeansLength_fromEnergySpec(density, energy)





def JeansMass_fromEnergySpec(density, energy):
  """
  Jeans Mass
  density : density
  energy  : specific energy  
  """
  Cs = SoundSpeed_fromEnergySpec(density,energy)
  return (np.pi**(5 / 2.) * Cs**3) / (6 * c.G**(3 / 2.) * np.sqrt(density))


def JeansMass_fromTemperature(density, temperature):
  """
  Jeans Mass
  density : density
  temperature  : temperature 
  """
  energy = Energy_fromTemperature(temperature)  
  return JeansMass_fromEnergySpec(density, energy)

















###########################################################################################
#  package:   RT
#  file:      blackbody.py
#  brief:     blackbody usefull routines
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Mladen Ivkovic <mladen.ivkovic@durham.ac.uk>
#
# This file is part of pNbody.
###########################################################################################

# ------------------------------------------
# Blackbody spectrum related functions.
# ------------------------------------------

import numpy as np


def B_nu(nu, T, kB, h_planck, c):
    """
    Return the blackbody energy density for
    a temperature T and frequency nu
    """
    res = 2.0 * h_planck * nu ** 3 / c ** 2 / (np.exp(h_planck * nu / kB / T) - 1.0)
    return res


def B_l(l, T, kB, h_planck, c):
    """
    Return the blackbody energy density for
    a temperature T and wavelength l
    """
    res = 2.0 * h_planck * c ** 2 / l ** 5 / (np.exp(h_planck * c / l / kB / T) - 1.0)
    return res




def B_nu_over_h_nu(nu, T, kB, h_planck, c):
    return B_nu(nu, T, kB, h_planck, c) / (h_planck * nu)


def nu_peak(T, kB, h_planck):
    """
    Return the (approximate) frequency where the peak of the
    blackbody energy density spectrum should be
    """
    return 2.82144 * kB * T / h_planck






#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libmiyamoto.py
#  brief:     Defines Miyamoto profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np


# miyam.f
# 
#       DO 60 I = 1, NTOT
#         R2 = X(I)**2 + Y(I)**2 + EPS
#         R = SQRT(R2)
#         C2 = Z(I)**2 + B12
#         C  = SQRT(C2)
#         D1 = A1 + C
#         D12 = D1*D1
#         D2 = A2 + C
#         D22 = D2*D2
# C
#         S12 = R2 + D12
#         S1 = SQRT(S12)
#         S22 = R2 + D22
#         S2 = SQRT(S22)
# C
#         T11 = GM1/S1
#         T13 = T11/S12
#         TD13 = T13*D1
#         TD15 = TD13/S12
#         T21 = GM2/S2
#         T23 = T21/S22
#         TD23 = T23*D2
#         TD25 = TD23/S22
# C P4RC2B2 : 4*Pi*Rho*C2/B2
#         P4RC2B2 = ( GM1*(A1*R2+(D1+C+C)*D12)/(S12*S12*S1) +
#      &              GM2*(A2*R2+(D2+C+C)*D22)/(S22*S22*S2) ) / C
#         SV = SQRT( .5/P4RC2B2 ) * (TD13 + TD23)
#         VT12 = R2*(T13+T23 - 3./P4RC2B2*(TD13+TD23)*(TD15+TD25) )
#         IF (VT12 .LT. 0.) THEN
#           NOUT = NOUT + 1
#           VV2 = 3.*SV*SV + VT12
#           VT1 = 0.
#           SV = SQRT(VV2/3.)
#         ELSE
#           VT1 = SQRT(VT12)
#         ENDIF
# C
# 	VT = SV*GASDEV() + VT1
#         VR = SV*GASDEV()
# C
#         PX(I) = (VR*X(I) - VT*Y(I))/R
#         PY(I) = (VR*Y(I) + VT*X(I))/R
#         PZ(I) = SV*GASDEV()
# C
#    60 CONTINUE



def Potential(M, a, b, R, z, G=1.):
    """
    Miyamoto-Nagai Potential
    """
    return -G * M / np.sqrt(R**2 + (a + np.sqrt(z**2 + b**2))**2)


def Vcirc(M, a, b, R, G=1.):
    """
    Miyamoto-Nagai Circular velocity
    """
    return np.sqrt(G * M / (R**2 + (a + b)**2)**1.5 * R**2)


def Omega(M, a, b, R, G=1.):
    """
    Miyamoto-Nagai Omega
    """
    return np.sqrt(G * M / (R**2 + (a + b)**2)**1.5)


def Kappa(M, a, b, R, G=1.):
    """
    Miyamoto-Nagai Kappa
    """

    r2 = R * R
    x = np.sqrt(r2 + (a + b)**2)
    return np.sqrt(G * M * (-3. * r2 / x**5 + 4. / x**3))


def Density(M, a, b, R, z, G=1.):
    """
    Miyamoto-Nagai Density
    """

    zb = np.sqrt(z**2 + b**2)
    azb2 = (a + zb)**2
    R2 = R * R

    cte = (b * b * M) / (4 * np.pi)
    return cte * (a * R2 + (a + 3 * zb) * azb2) / ((R2 + azb2)**2.5 * zb**3)


def dR_Potential(M, a, b, R, z, G=1.):
    """
    first derivative in R
    """

    R2 = R * R
    zb = np.sqrt(z**2 + b**2)
    azb2 = (a + zb)**2

    R2azb2 = R2 + azb2

    return G * M * R / R2azb2**(3.0 / 2.0)


def d2R_Potential(M, a, b, R, z, G=1.):
    """
    second derivative in R
    """

    R2 = R * R
    zb = np.sqrt(z**2 + b**2)
    azb2 = (a + zb)**2

    R2azb2 = R2 + azb2

    return -3 * G * M * R2 * \
        R2azb2**(-5.0 / 2.0) + G * M * R2azb2**(-3.0 / 2.0)


def dz_Potential(M, a, b, R, z, G=1.):
    """
    first derivative in R
    """

    R2 = R * R
    zb = np.sqrt(z**2 + b**2)
    azb2 = (a + zb)**2

    R2azb2 = R2 + azb2

    return G * M * z * (a + zb) / (R2azb2**(3.0 / 2.0) * zb)


def d2z_Potential(M, a, b, R, z, G=1.):
    """
    second derivative in R
    """

    R2 = R * R
    z2 = z * z
    zb2 = z**2 + b**2
    zb = np.sqrt(zb2)
    azb = a + zb
    azb2 = azb**2

    R2azb2 = R2 + azb2

    c1 = -3 * z2 * azb2 / (R2azb2**(5.0 / 2.0) * zb2)
    c2 = z2 / (R2azb2**(3.0 / 2.0) * zb2)
    c3 = -  z2 * azb / (R2azb2**(3.0 / 2.0) * zb**3.0)
    c4 = azb / (R2azb2**(3.0 / 2.0) * zb)

    return G * M * (c1 + c2 + c3 + c4)


def Sigma_z(M, a, b, R, z, G=1.):
    """
    Return sigma_z from Jeans equation : 1/rho Int( rho * dzPhi * dz )
    """

    R2 = R * R
    b2 = b * b
    z2 = z * z
    c2 = z2 + b2
    c = np.sqrt(c2)

    zb2 = z**2 + b**2
    zb = np.sqrt(zb2)
    azb = a + zb
    azb2 = azb**2
    R2azb2 = R2 + azb2

    TD = G * M * (a + c) / R2azb2**(3.0 / 2.0)

    return np.sqrt(b2 / (8 * np.pi * c2) * TD**2 / Density(G, M, a, b, R, z))


def Sigma_zbis(M, a, b, R, z, G=1.):
    """
    Same than Sigma_z, but optimized
    """

    GM1 = G * M
    A1 = a
    B1 = b

    B12 = B1 * B1
    R2 = R * R
    C2 = z**2 + B12
    C = np.sqrt(C2)
    D1 = A1 + C
    D12 = D1 * D1

    S12 = R2 + D12
    S1 = np.sqrt(S12)

    T11 = GM1 / S1
    T13 = T11 / S12
    TD13 = T13 * D1
    TD15 = TD13 / S12

    P4RC2B2 = (GM1 * (A1 * R2 + (D1 + C + C) * D12) / (S12 * S12 * S1)) / C

    SV = np.sqrt(0.5 / P4RC2B2) * TD13

    return SV


def Sigma_t(M, a, b, R, z, G=1.):
    """
    Return sigma_z from Jeans equation : 1/rho Int( rho * dzPhi * dz )

    sigma_t^2 =   R*d/dr(Phi) + R/rho*d/dr(rho*sigma_z^2)

    """

    GM1 = G * M
    A1 = a
    B1 = b
    B12 = b * b

    R2 = R * R
    C2 = z**2 + B12
    C = np.sqrt(C2)
    D1 = A1 + C
    D12 = D1 * D1

    S12 = R2 + D12
    S1 = np.sqrt(S12)

    T11 = GM1 / S1
    T13 = T11 / S12
    TD13 = T13 * D1
    TD15 = TD13 / S12

    P4RC2B2 = (GM1 * (A1 * R2 + (D1 + C + C) * D12) / (S12 * S12 * S1)) / C

    SV = np.sqrt(.5 / P4RC2B2) * TD13
    VT12 = R2 * (T13 - 3. / P4RC2B2 * TD13 * TD15)

    if (VT12 >= 0.):
        VT1 = np.sqrt(VT12)
    else:
        VT1 = 0.

    return VT1


def SurfaceDensity(M, a, b, R, G=1.):
    """
    Miyamoto-Nagai Surface density
    """

    z = 0
    dz = b / 100.
    integral = 0.0
    err = 1.
    i = 0

    while (i < 300):
        integral = integral + Density(G, M, a, b, R, z)
        z = z + dz

        #err = abs(old_integral-integral)/integral

        i = i + 1

    return integral * 2. * dz



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      Movie.py
#  brief:     Movie functions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


import string

import numpy as np

from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from PIL import ImagePalette
from PIL import ImageFont


class Movie:
    """
    a Movie class
    """

    ##########################
    def __init__(self, name, mode=None):
        ##########################

        self.name = name
        self.mode = mode
        self.starttime = 0.0
        self.stoptime = 0.0
        self.dt = 0.0
        self.npic = 0
        self.numByte = 0
        self.numLine = 0
        self.headerlength = 0
        self.blocksize = 0
        self.filmsize = 0
        self.shape = (0, 0)

        self.current_npic = 0

        # self.film

    ##########################

    def open(self, mode='rb', readall=0):
        ##########################
        """
        open a file
        """

        numByte = 256

        # open the file
        self.film = open(self.name, mode)

        ###############################################
        # read the header and find numByte and numLine
        ###############################################

        self.header = self.film.read(256)

        if len(self.header) != 256:
            raise Exception("film length < 256, ,stop")

        self.film.seek(240)

        str1 = self.film.read(8)
        str2 = self.film.read(8)

        if (str1 != "	" and str2 != "        "):
            self.numByte = int(str1)
            self.numLine = int(str2)
        else:
            self.numByte = 256
            self.numLine = 192

        if self.mode == 'RGB':
            self.numLine = self.numLine * 3

        if self.numByte < 256:
            self.headerlength = 256
        else:
            self.headerlength = self.numByte

        ###############################################
        # read the header and find numByte and numLine
        ###############################################

        self.blocksize = int(self.numLine * self.numByte + self.headerlength)
        self.shape = (self.numByte, self.numLine)

        #############################
        # check film length
        #############################

        self.film.seek(0, 2)
        self.filmsize = self.film.tell()
        self.film.seek(0)

        np.fmod(self.filmsize, self.blocksize)

        self.npic = self.filmsize // self.blocksize

        #############################
        # read times
        #############################

        self.film.seek(0)
        self.starttime = float(self.film.read(self.headerlength)[:8])
        self.film.seek(0)

        self.moveto(self.npic)
        self.stoptime = float(self.film.read(self.headerlength)[:8])
        self.film.seek(0)

        if self.npic > 1:
            self.dt = (self.stoptime - self.starttime) / self.npic
        else:
            self.dt = 0

        self.current_npic = 0


    ##########################
    def info(self):
        ##########################
        """
        give info
        """

        print(("""
    -- %s  --

    initial time	    : %8.3f
    final   time	    : %8.3f
    dt  		    : %8.3f
    number of images	    : %d
    number of collumns      : %d
    number of lines	    : %d
    length of header	    : %d
    length of block         : %d
    length of film          : %d

    current_npic            : %d
    """ % (self.name, self.starttime, self.stoptime, self.dt, self.npic, self.numByte, self.numLine, self.headerlength, self.blocksize, self.filmsize, self.current_npic)))

    ##########################

    def new(self, numByte, numLine):
        ##########################

        self.numByte = numByte
        self.numLine = numLine

        if self.numByte < 256:
            self.headerlength = 256
        else:
            self.headerlength = self.numByte

        self.film = open(self.name, "wb")

    ##########################
    def close(self):
        ##########################

        self.film.close()

    ##########################
    def read_one(self, mode=None):
        ##########################

        # reading next label
        time = self.film.read(self.headerlength)

        if len(time) == self.headerlength:

            # record starttime
            self.current_time = float(time[:8])

            # reading next data
            data = self.film.read(self.numByte * self.numLine)
            if len(data) == self.numByte * self.numLine:

                self.current_image = data

                if mode == "array":
                    return np.reshape(np.fromstring(data, np.uint8),
                                   (self.numLine, self.numByte))
                elif mode == "image":
                    return Image.frombytes("P", 
                        (self.numByte, self.numLine), data)
                else:
                    return data

    ##########################
    def read_one_with_time(self, mode=None):
        ##########################

        # reading next label
        time = self.film.read(self.headerlength)

        if len(time) == self.headerlength:

            # record starttime
            self.current_time = float(time[:8])

            shape = (self.numByte, self.numLine)

            if mode == "array":
                data = self.film.read(self.numByte * self.numLine)
                return time, np.reshape(
                    np.fromstring(
                        data, 'b'), (self.numLine, self.numByte))

            elif mode == "image":
                data = self.film.read(self.numByte * self.numLine)
                return time, Image.frombytes("P", shape, data)

            elif mode == "image_rgb":

                dataR = self.film.read(self.numByte * self.numLine)
                self.film.read(self.headerlength)
                dataG = self.film.read(self.numByte * self.numLine)
                self.film.read(self.headerlength)
                dataB = self.film.read(self.numByte * self.numLine)

                imageR = Image.frombytes("L", shape, dataR)
                imageG = Image.frombytes("L", shape, dataG)
                imageB = Image.frombytes("L", shape, dataB)

                return time, Image.merge('RGB', (imageR, imageG, imageB))

            else:
                data = self.film.read(self.numByte * self.numLine)
                return time, data

        else:
            return None, None

    ##########################
    def read(self, skip=0, mode='array'):
        ##########################
        """
        skip =  0 	: read image at the current position
        skip =  1	: skip an image
        skip = -1   : read the image before (go back)
        skip = -2   : skip an image before  (go back)
        """

        # move relative to the current position
        try:
            self.film.seek(skip * self.blocksize, 1)
        except IOError:
            self.moveto(0)
            return -1, 0, -1

        self.current_npic = self.film.tell() // self.blocksize

        if self.current_npic > self.npic - 1:
            self.moveto(self.npic)
            return -2, self.current_npic, -2

        # reading next label
        time = self.film.read(self.headerlength)

        # record starttime
        self.current_time = float(time[:8])

        if mode == "array":
            data = self.film.read(self.numByte * self.numLine)
            return self.current_time, self.current_npic, np.transpose(
                np.reshape(np.fromstring(data, 'b'), (self.numLine, self.numByte)))

        elif mode == "image":
            data = self.film.read(self.numByte * self.numLine)
            return self.current_time, self.current_npic, Image.frombytes(
                "P", self.shape, data)

        elif mode == "image_rgb":

            dataR = self.film.read(self.numByte * self.numLine)
            self.film.read(self.headerlength)
            dataG = self.film.read(self.numByte * self.numLine)
            self.film.read(self.headerlength)
            dataB = self.film.read(self.numByte * self.numLine)

            imageR = Image.frombytes("L", self.shape, dataR)
            imageG = Image.frombytes("L", self.shape, dataG)
            imageB = Image.frombytes("L", self.shape, dataB)

            return self.current_time, self.current_npic, Image.merge(
                'RGB', (imageR, imageG, imageB))

        else:
            data = self.film.read(self.numByte * self.numLine)
            return self.current_time, self.current_npic, data

    ##########################
    def moveto(self, npic):
        ##########################

        npic = min(npic, self.npic - 1)
        npic = max(npic, 0)

        current_npic = self.film.tell() // self.blocksize
        dnpic = npic - current_npic
        db = dnpic * self.blocksize
        self.film.seek(db, 1)

        self.current_npic = self.film.tell() // self.blocksize

    ###########################
    def write_pic(self, time, data):
        ###########################

        recsize = self.numByte

        record = b'%8.3f' % time

        s1 = bytes.ljust(record, 240)
        s2 = bytes.ljust(bytes(self.numByte), 8)
        if self.mode == 'RGB':
            s3 = bytes.ljust(bytes(self.numLine // 3), 8)
        else:
            s3 = bytes.ljust(bytes(self.numLine), 8)
        record = s1 + s2 + s3
        record = bytes.ljust(record, recsize)

        self.film.write(record)
        self.film.write(data)

    ###########################
    def write_pic_rgb(self, time, dataR, dataG, dataB):
        ###########################

        recsize = self.numByte

        record = '%8.3f' % time

        s1 = str.ljust(record, 240)
        s2 = str.ljust(repr(self.numByte), 8)
        s3 = str.ljust(repr(self.numLine), 8)
        record = s1 + s2 + s3
        record = str.ljust(record, recsize)

        self.film.write(record)
        self.film.write(dataR)
        self.film.write(dataG)
        self.film.write(dataB)

    ###########################
    def get_img(self, data):
        ###########################
        """
        can be replaced by read_one with option "image"
        """
        shape = (self.numByte, self.numLine)
        image = Image.frombytes("P", shape, data)
        return image


###################################
def append_h(numByte, numLine, datas):
    ###################################

    newdata = b''

    # loop over the lines
    for j in range(numLine):

        j1 = j * numByte 		# set cursors
        j2 = j1 + numByte

        # loop over the images
        for data in datas:

            newdata = newdata + data[j1:j2]  # simply sum

    return newdata



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      mpiwrapper.py
#  brief:     Defines mpi routines for pNbody
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################



import numpy as np
from . import libutil
import sys


class myMPI:
    def __init__(self):
        pass


try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    comm_self = MPI.COMM_SELF
    ThisTask = comm.Get_rank()
    NTask = comm.Get_size()
    Procnm = MPI.Get_processor_name()
    Rank = ThisTask
except BaseException:
    MPI = myMPI()
    MPI.sum = np.sum
    MPI.SUM = MPI.sum
    ThisTask = 0
    NTask = 1
    Rank = ThisTask
    Procnm = ""


def mpi_IsMaster():
    return ThisTask == 0


def mpi_Rank():
    return ThisTask


def mpi_ThisTask():
    return ThisTask


def mpi_NTask():
    return NTask


def mpi_barrier():
    if NTask > 1:
        comm.barrier()
    else:
        pass

#####################################
#
# Output functions
#
#####################################


def mpi_iprint(*obj):
    """
    Print, including info on node.
    """
    print("[%d] : " % ThisTask, *obj)


def mpi_pprint(*obj):
    """
    Synchronized print.
    """
    if NTask > 1:
        comm.barrier()
    print(*obj)


def mpi_rprint(*obj):
    """
    Rooted print.
    """
    if mpi_IsMaster():
        print(*obj)


#####################################
#
# communication functions
#
#####################################


def mpi_send(x, dest):
    """
    Send x to node dest.
    When there is only one is defined, it does nothing.
    """

    if NTask > 1:
        return comm.send(x, dest=dest)
    else:
        pass


def mpi_recv(source):
    """
    Return a variable sent by node \\var{source}.
    When there is only one is defined, it does nothing.

    """
    if NTask > 1:
        return comm.recv(source=source)
    else:
        pass


def mpi_sendrecv(x, dest, source):
    """
    """
    if NTask > 1:
        return comm.sendrecv(x, dest=dest, source=source)
    else:
        pass


def mpi_reduce(x, root=0, op=MPI.SUM):
    """
    Reduce x from all node only for root.
    When there is only one is defined, the function return x.
    """

    if NTask > 1:
        sx = comm.reduce(x, op=op, root=root)
    else:
        sx = x

    return sx


def mpi_allreduce(x, op=MPI.SUM):
    """
    Reduce x from all node for all nodes.
    When there is only one is defined, the function return x.
    """

    if NTask > 1:
        sx = comm.allreduce(x, op=op)
    else:
        sx = x

    return sx


def mpi_bcast(x, root=0):
    """
    Broadcast from node root the variable x.
    When there is only one is defined, it simplay returns x.
    """

    if NTask > 1:
        x = comm.bcast(x, root=root)
    else:
        x = x

    return x


def mpi_gather(x, root=0):
    """
    Gather x from all nodes to node dest.
    Returns a list.
    """

    if NTask > 1:
        x = comm.gather(x, root=root)
    else:
        x = [x]

    return x


def mpi_allgather(x):
    """
    Gather x from all to all.
    Returns a list.
    """

    if NTask > 1:
        x = comm.allgather(x)
    else:
        x = [x]

    return x


def mpi_AllgatherAndConcatArray(vec):
    """
    AllGather array vec and concatenate it in a unique array
    (concatenation order is reversed).
    """

    if NTask > 1:

        vec_all = np.array([], vec.dtype)
        if vec.ndim == 1:
            vec_all.shape = (0,)
        else:
            vec_all.shape = (0, vec.shape[1])

        if ThisTask == 0:

            for Task in range(NTask - 1, -1, -1):

                if Task != 0:
                    v = comm.recv(source=Task)
                    vec_all = np.concatenate((vec_all, v))
                else:
                    vec_all = np.concatenate((vec_all, vec))

        else:
            comm.send(vec, dest=0)

        # send to everybody
        xi = comm.bcast(vec_all)

        return xi

    else:
        return vec


#####################################
#
# array functions
#
#####################################


def mpi_sum(x):
    """
    Sum elements of array x.
    """

    if NTask > 1:
        sx = comm.allreduce(np.sum(x), op=MPI.SUM)
    else:
        sx = np.sum(x)

    return sx


def mpi_min(x):
    """
    Minimum element of array x.
    """

    if NTask > 1:
        mx = comm.allreduce(min(x), op=MPI.MIN)
    else:
        mx = min(x)

    return mx


def mpi_max(x):
    """
    Maximum element of array x.
    """

    if NTask > 1:
        mx = comm.allreduce(max(x), op=MPI.MAX)
    else:
        mx = max(x)

    return mx


def mpi_mean(x):
    """
    Mean of elements of array x.
    """

    if NTask > 1:
        sm = comm.allreduce(np.sum(x), op=MPI.SUM)
        sn = comm.allreduce(len(x), op=MPI.SUM)
        mn = sm / sn
    else:
        mn = x.mean()

    return mn


def mpi_len(x):
    """
    Length of array x.
    """

    if NTask > 1:
        ln = comm.allreduce(len(x), op=MPI.SUM)
    else:
        ln = len(x)

    return ln


def mpi_arange(n):
    """
    Create an integer array containing elements from 0 to n
    spread over all nodes.
    """

    if NTask > 1:
        ns = np.array(comm.allgather(n))
        c = ((NTask - 1 - ThisTask) > np.arange(len(ns) - 1, -1, -1))
        i = np.sum(ns * c)
        v = np.arange(i, i + ns[ThisTask])
    else:
        v = np.arange(n)

    return v


def mpi_sarange(npart_all):
    """
    Create an integer array containing elements from 0 to n,
    spreaded over all nodes. The repartition of elements and
    type of elements over nodes is given by the array npart_all

    """

    npart = npart_all[ThisTask]

    if NTask > 1:
        v = np.array([], int)
        o = 0

        for i in np.arange(len(npart)):

            # number of particles of this type for this proc
            n = npart_all[ThisTask, i]
            n_tot = np.sum(npart_all[:, i])

            ns = np.array(comm.allgather(n))
            c = ((NTask - 1 - ThisTask) > np.arange(len(ns) - 1, -1, -1))
            j = np.sum(ns * c)
            vj = np.arange(o + j, o + j + ns[ThisTask])
            v = np.concatenate((v, vj))

            o = o + n_tot

    else:
        n = np.sum(np.ravel(npart))
        v = np.arange(n)

    return v


def mpi_argmax(x):
    """
    Find the arument of the amximum value in x.

    idx = (p,i)  : where i = index in proc p

    """

    if NTask > 1:
        # 1) get all max and all argmax
        maxs = np.array(comm.allgather(max(x)))
        args = np.array(comm.allgather(np.argmax(x)))

        # 2) choose the right one
        p = np.argmax(maxs)	  # proc number
        i = args[p] 	  # index in proc p

    else:
        p = 0
        i = np.argmax(x)

    return p, i


def mpi_argmin(x):
    """
    Find the arument of the maximum value in x.
    idx = (p,i)  : where i = index in proc p

    """

    if NTask > 1:
        # 1) get all mins and all argmins
        mins = np.array(comm.allgather(min(x)))
        args = np.array(comm.allgather(np.argmin(x)))

        # 2) choose the right one
        p = np.argmin(mins)  # proc number
        i = args[p]		# index in proc p

    else:
        p = 0
        i = np.argmin(x)

    return p, i


def mpi_getval(x, idx):
    """
    Return the value of array x corresponding to the index idx.

    idx = (p,i)  : where i = index in proc p
    equivalent to x[i] from proc p

    """

    p = idx[0]
    i = idx[1]

    if NTask > 1:
        # send to everybody
        xi = comm.bcast(x[i], root=p)

    else:
        xi = x[i]

    return xi


def mpi_histogram(x, bins):
    """
    Return an histogram of vector x binned using binx.
    """

    # histogram
    n = np.searchsorted(np.sort(x), bins)
    n = np.concatenate([n, [len(x)]])
    hx = n[1:] - n[:-1]

    if NTask > 1:
        hx = comm.allreduce(hx, op=MPI.SUM)

    return hx


#####################################
#
# exchange function
#
#####################################


#################################
def mpi_find_a_toTask(begTask, fromTask, ex_table, delta_n):
    #################################
    """
    This function is used to find recursively an exange table
    """
    for toTask in range(begTask, NTask):
        if delta_n[toTask] > 0:		# find a node who accept particles

            if delta_n[toTask] + delta_n[fromTask] >= 0:  # can give all
                delta = -delta_n[fromTask]
            else:						# can give only a fiew
                delta = +delta_n[toTask]

            ex_table[fromTask, toTask] = +delta
            ex_table[toTask, fromTask] = -delta

            delta_n[fromTask] = delta_n[fromTask] + delta
            delta_n[toTask] = delta_n[toTask] - delta

            if delta_n[fromTask] == 0:  # everything has been distributed
                return ex_table
            else:
                return mpi_find_a_toTask(
                    begTask + 1, fromTask, ex_table, delta_n)


#################################
def mpi_GetExchangeTable(n_i):
    #################################
    """
    This function returns the exchange table
    """

    # n_i : initial repartition
    ntot = np.sum(n_i)
    if ntot == 0:
        return np.zeros((NTask, NTask))

    # final repartition
    n_f = np.zeros(NTask)
    for Task in range(NTask - 1, -1, -1):
        n_f[Task] = ntot / NTask + ntot % NTask * (Task == 0)

    # delta
    delta_n = n_f - n_i

    # find who gives what to who
    ex_table = np.zeros((NTask, NTask))

    for fromTask in range(NTask):
        if delta_n[fromTask] < 0:		  # the node gives particles
            ex_table = mpi_find_a_toTask(0, fromTask, ex_table, delta_n)

    return ex_table


#################################
def mpi_ExchangeFromTable(T, procs, ids, vec, num):
    #################################
    """
    Exchange an array according to a transfer array T

    T : exchange table

    procs : list of processor     (from Tree.GetExchanges())
    ids   : list of id		(from Tree.GetExchanges())

    vec   : vector to exchange
    num   : id corresponding to particles
    """

    # now, we have to send / recv
    SendPart = T[NTask - 1 - ThisTask, :]
    RecvPart = T[:, ThisTask]

    new_procs = np.array([])
    new_vec = np.array([])

    if vec.ndim == 1:
        #new_vec.shape = (0,3)
        pass
    elif vec.ndim == 2:
        new_vec.shape = (0, 3)

    # send/recv (1)
    for i in range(NTask):

        if i > ThisTask:
            # here, we send to i

            N = T[NTask - 1 - ThisTask, i]
            comm.send(N, i)

            if N > 0:
                to_procs = np.compress((procs == i), procs, axis=0)
                to_ids = np.compress((procs == i), ids, axis=0)
                to_vec = libutil.compress_from_lst(vec, num, to_ids)
                comm.send(to_vec, i)

                vec = libutil.compress_from_lst(vec, num, to_ids, reject=True)
                num = libutil.compress_from_lst(num, num, to_ids, reject=True)

        elif i < ThisTask:
            N = comm.recv(i)
            if N > 0:
                new_vec = np.concatenate((new_vec, comm.recv(i)))

    # send/recv (1)
    for i in range(NTask):

        if i < ThisTask:
            # here, we send to i

            N = T[NTask - 1 - ThisTask, i]
            comm.send(N, i)

            if N > 0:
                to_procs = np.compress((procs == i), procs, axis=0)
                to_ids = np.compress((procs == i), ids, axis=0)
                to_vec = libutil.compress_from_lst(vec, num, to_ids)
                comm.send(to_vec, i)

                vec = libutil.compress_from_lst(vec, num, to_ids, reject=True)
                num = libutil.compress_from_lst(num, num, to_ids, reject=True)

        elif i > ThisTask:
            N = comm.recv(i)
            if N > 0:
                new_vec = np.concatenate((new_vec, comm.recv(i)))

    # check
    c = (new_procs != ThisTask).astype(int)
    if np.sum(c) > 0:
        print("here we are in trouble")
        sys.exit()

    return np.concatenate((vec, new_vec))

    # !!!!!!!!!!!!!!!!!!!!!!! this has to be changed ....


#####################################
#
# io functions
#
#####################################


#################################
def mpi_ReadAndSendBlock(
        f,
        data_type,
        shape=None,
        byteorder=sys.byteorder,
        split=None,
        htype=np.int32):
    #################################
    """

    Read and broadcast a binary block.

    data_type = int,float32,float
    or
    data_type = array

    shape     = tuple
    """

    #####################
    # read first header
    #####################
    if ThisTask == 0:

        try:
            nb1 = np.frombuffer(f.read(4), htype)
            if sys.byteorder != byteorder:
                nb1.byteswap(True)
            nb1 = nb1[0]

        except IndexError:
            raise Exception("ReadBlockError")

    #####################
    # read a tuple
    #####################
    if isinstance(data_type, tuple):

        if ThisTask == 0:

            data = []
            for tpe in data_type:

                if isinstance(tpe, int):
                    val = f.read(tpe)
                else:
                    bytes = np.dtype(tpe).itemsize
                    val = np.frombuffer(f.read(bytes), tpe)
                    if sys.byteorder != byteorder:
                        val.byteswap(True)

                    val = val[0]

                data.append(val)

        # send the data
        if NTask > 1:

            if ThisTask == 0:
                for Task in range(1, NTask):
                    comm.send(data, dest=Task)

            else:
                data = comm.recv(source=0)

    #####################
    # read an array
    #####################
    else:

        if split:
            if ThisTask == 0:

                bytes_per_elt = data_type.bytes

                if shape is not None:
                    ndim = shape[1]
                else:
                    ndim = 1

                nelt = nb1 / bytes_per_elt / ndim
                nleft = nelt

                nread = nelt / NTask

                for Task in range(NTask - 1, -1, -1):

                    if nleft < 2 * nread and nleft > nread:
                        nread = nleft

                    nleft = nleft - nread
                    data = f.read(nread * bytes_per_elt * ndim)
                    shape = (nread, ndim)

                    if Task == ThisTask:			# this should be last
                        data = np.frombuffer(data, data_type)
                        if sys.byteorder != byteorder:
                            data.byteswap(True)
                        data.shape = shape
                    else:
                        comm.send(data, dest=Task)
                        comm.send(shape, dest=Task)

            else:
                data = comm.recv(source=0)
                shape = comm.recv(source=0)
                data = np.frombuffer(data, data_type)

                if sys.byteorder != byteorder:
                    data.byteswap(True)

                data.shape = shape

        else:

            if ThisTask == 0:

                data = np.frombuffer(f.read(nb1), data_type)
                if sys.byteorder != byteorder:
                    data.byteswap(True)

            # send the data
            if NTask > 1:
                if ThisTask == 0:
                    for Task in range(1, NTask):
                        comm.send(data, dest=Task)
                else:
                    data = comm.recv(source=0)

    #####################
    # read second header
    #####################
    if ThisTask == 0:
        nb2 = np.frombuffer(f.read(4), htype)
        if sys.byteorder != byteorder:
            nb2.byteswap(True)
        nb2 = nb2[0]

        if nb1 != nb2:
            raise Exception("ReadBlockError")

    # reshape if needed
    if split:
        # already done before
        pass
    else:
        if shape is not None:
            data.shape = shape

    return data

#################################


def mpi_ReadAndSendArray(
        f,
        data_type,
        shape=None,
        byteorder=sys.byteorder,
        npart=None,
        skip=False,
        htype=np.int32):
    #################################
    """
    Read and Brodcast a binary block assuming it contains an array.
    """

    oneD = False

    if shape is not None:
        if len(shape) == 1:
            shape = (shape[0], 1)
            oneD = True

        n_dim = shape[1]
        data = np.array([], data_type)
        data.shape = (0, shape[1])
        n_elts = shape[0]
    else:
        raise Exception(
            "mpi_ReadAndSendArray : var shape must be defined here.")
        #n_dim = 1
        #data = np.array([],data_type)
        #data.shape = (0,)

    nbytes_per_elt = np.dtype(data_type).itemsize * n_dim
    n_left = nbytes_per_elt * n_elts

    # check npart

    if npart is not None:
        ntype = len(npart)
    else:
        ntype = 1
        npart = np.array([n_elts])

    # this may be wrong in 64b, as nb1 is not equal to the number of bytes in block
    # if np.sum(npart,0) != n_elts:
    #  raise "We are in trouble here : np.sum(npart)=%d n_elts=%d"%(np.sum(npart,0),n_elts)

    #####################
    # read first header
    #####################
    if ThisTask == 0:

        try:
            nb1 = np.frombuffer(f.read(4), htype)
            if sys.byteorder != byteorder:
                nb1.byteswap(True)
            nb1 = nb1[0]

            # this may be wrong in 64b, as nb1 is not equal to the number of bytes in block
            # if nb1 != n_left:
            #  raise "We are in trouble here : nb1=%d n_left=%d"%(nb1,n_left)

        except IndexError:
            raise Exception("ReadBlockError")

    #####################
    # read data
    #####################

    if ThisTask == 0:

        if skip:
            nbytes = np.sum(npart) * nbytes_per_elt
            print(("  skipping %d bytes... " % (nbytes)))
            f.seek(nbytes, 1)
            data = None

        else:
            for i in range(ntype):

                n_write = libutil.get_n_per_task(npart[i], NTask)

                for Task in range(NTask - 1, -1, -1):

                    sdata = f.read(nbytes_per_elt * n_write[Task])

                    # send block for a slave
                    if Task != 0:
                        comm.send(sdata, dest=Task)
                    # keep block for the master
                    else:
                        sdata = np.frombuffer(sdata, data_type)
                        if shape is not None:
                            sdata.shape = (n_write[Task], shape[1])
                        data = np.concatenate((data, sdata))

    else:
        if skip:
            data = None
        else:
            for i in range(ntype):
                sdata = comm.recv(source=0)
                sdata = np.frombuffer(sdata, data_type)
                if shape is not None:
                    ne = len(sdata) // shape[1]
                    sdata.shape = (ne, shape[1])

                data = np.concatenate((data, sdata))

    if oneD and data is not None:
        data.shape = (data.shape[0],)

    #####################
    # read second header
    #####################
    if ThisTask == 0:
        nb2 = np.frombuffer(f.read(4), htype)
        if sys.byteorder != byteorder:
            nb2.byteswap(True)
        nb2 = nb2[0]

        if nb1 != nb2:
            raise Exception("ReadBlockError")

    return data


#################################
def mpi_GatherAndWriteArray(
        f,
        data,
        byteorder=sys.byteorder,
        npart=None,
        htype=np.int32):
    #################################
    """
    Gather and array and write it in a binary block.

    data = array

    shape     = tuple
    """

    # check npart
    if npart is not None:
        pass
    else:
        npart = np.array([len(data)])

    data_nbytes = mpi_allreduce(data.nbytes)

    if ThisTask == 0:

        nbytes = np.array([data_nbytes], htype)

        if sys.byteorder != byteorder:
            nbytes.byteswap(True)

        # write header 1
        f.write(nbytes.tostring())

    # loop over particles type

    for i in range(len(npart)):

        n_write = libutil.get_n_per_task(npart[i], NTask)

        if ThisTask == 0:

            for Task in range(NTask - 1, -1, -1):

                if Task != 0:
                    sdata = comm.recv(source=Task)
                else:
                    i1 = int(np.sum(npart[:i]))
                    i2 = int(i1 + npart[i])
                    sdata = data[i1:i2]

                if sys.byteorder != byteorder:
                    sdata.byteswap(True)

                f.write(sdata.tostring())

        else:

            i1 = int(np.sum(npart[:i]))
            i2 = int(i1 + npart[i])

            comm.send(data[i1:i2], dest=0)

    if ThisTask == 0:
        # write header 2
        f.write(nbytes.tostring())


#################################
def mpi_OldReadAndSendArray(
        f,
        data_type,
        shape=None,
        skip=None,
        byteorder=sys.byteorder,
        nlocal=None):
    #################################
    """

    Read and Brodcast a binary block assuming it contains an array.
    The array is splitted acroding to the variable nlocal.

    data_type = array type
    shape     = tuple

    nlocal : array   NTask x Npart
             array   NTask

    """

    #####################
    # read first header
    #####################
    if ThisTask == 0:

        try:
            nb1 = np.frombuffer(f.read(4), np.int32)
            if sys.byteorder != byteorder:
                nb1.byteswap(True)
            nb1 = nb1[0]

        except IndexError:
            raise Exception("ReadBlockError")

    #####################
    # read an array
    #####################

    if nlocal.ndim != 2:
        raise Exception("OldReadAndSendArray error", "nlocal must be of rank 2")
    else:
        ntype = nlocal.shape[1]

    if shape is not None:
        ndim = shape[1]
        data = np.array([], data_type)
        data.shape = (0, shape[1])
    else:
        ndim = 1
        data = np.array([], data_type)
        data.shape = (0,)

    nbytes_per_elt = data_type.itemsize * ndim

    if ThisTask == 0:

        for i in range(ntype):
            for Task in range(NTask - 1, -1, -1):
                sdata = f.read(nbytes_per_elt * int(nlocal[Task, i]))

                # send block for a slave
                if Task != 0:
                    comm.send(sdata, dest=Task)
                # keep block for the master
                else:
                    sdata = np.frombuffer(sdata, data_type)
                    if shape is not None:
                        sdata.shape = (nlocal[ThisTask, i], shape[1])
                    data = np.concatenate((data, sdata))

    else:

        for i in range(ntype):
            sdata = comm.recv(source=0)
            sdata = np.frombuffer(sdata, data_type)
            if shape is not None:
                sdata.shape = (nlocal[ThisTask, i], shape[1])

            data = np.concatenate((data, sdata))

    #####################
    # read second header
    #####################
    if ThisTask == 0:
        nb2 = np.frombuffer(f.read(4), np.int32)
        if sys.byteorder != byteorder:
            nb2.byteswap(True)
        nb2 = nb2[0]

        if nb1 != nb2:
            raise Exception("ReadBlockError")

    return data


#################################
def mpi_OldGatherAndWriteArray(f, data, byteorder=sys.byteorder, nlocal=None):
    #################################
    """
    Gather and array and write it in a binary block.

    data = array

    shape	    = tuple
    """

    if nlocal.ndim != 2:
        raise Exception("OldGatherAndWriteArray error", "nlocal must be of rank 2")
    else:
        ntype = nlocal.shape[1]

    data_size = mpi_allreduce(data.size)

    if ThisTask == 0:

        nbytes = np.array([data.dtype.itemsize * data_size], np.int32)

        if sys.byteorder != byteorder:
            nbytes.byteswap(True)

        # write header 1
        f.write(nbytes.tostring())

    # loop over particles type
    i1 = 0

    for i in range(ntype):

        if ThisTask == 0:

            for Task in range(NTask - 1, -1, -1):

                if Task != 0:
                    sdata = comm.recv(source=Task)
                else:

                    i2 = i1 + int(nlocal[Task, i])

                    sdata = data[i1:i2]
                    i1 = i2

                if sys.byteorder != byteorder:
                    sdata.byteswap(True)

                f.write(sdata.tostring())

        else:

            i2 = i1 + int(nlocal[ThisTask, i])

            comm.send(data[i1:i2], dest=0)
            i1 = i2

    if ThisTask == 0:
        # write header 2
        f.write(nbytes.tostring())



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      cooling.py
#  brief:     cooling functions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from pNbody import *

try:
    import pNbody.cooling_with_metals as libcooling
except BaseException:
    pass

"""
In pNbody we need to wrap cooling_with_metals in cooling
in order to initialize it safely with the default parameters.

This module is a wrapper around the defaut GEAR cooling function
using the metallicity.
"""



# THIS IS NOW DONE OUTSIDE
# as we cannot acess to nb.localsystem_of_units
#
# print "!!!!!! in cooling.py  !!!!"
# print "!!!!!! this is bad    !!!! we should never use UNITSPARAMETERFILE"
# print "!!!!!! this is bad    !!!! we should never use UNITSPARAMETERFILE"
#
# unitsparameters = param.Params(UNITSPARAMETERFILE,None)
# libcooling.init_cooling(unitsparameters.get_dic())
# print unitsparameters.get_dic()
# sys.exit()
#
# # set system of units
# params = {}
# params['UnitLength_in_cm']	   = unitsparameters.get('UnitLength_in_cm')
# params['UnitVelocity_in_cm_per_s'] = unitsparameters.get('UnitVelocity_in_cm_per_s')
# params['UnitMass_in_g'] 	   = unitsparameters.get('UnitMass_in_g')
# localsystem_of_units = units.Set_SystemUnits_From_Params(params)



init_cooling = libcooling.init_cooling

get_lambda_from_Density_Temperature_FeH = libcooling.get_lambda_from_Density_Temperature_FeH
get_lambda_from_Density_EnergyInt_FeH = libcooling.get_lambda_from_Density_EnergyInt_FeH
get_lambda_from_Density_Entropy_FeH = libcooling.get_lambda_from_Density_Entropy_FeH
get_lambda_normalized_from_Temperature_FeH = libcooling.get_lambda_normalized_from_Temperature_FeH
get_cooling_time_from_Density_Temperature_FeH = libcooling.get_cooling_time_from_Density_Temperature_FeH
get_cooling_time_from_Density_EnergyInt_FeH = libcooling.get_cooling_time_from_Density_EnergyInt_FeH


def check():

    print()
    libcooling.PrintParameters()

    rho = 3.0364363e-06
    u = 0.0015258887
    fe = 0.
    l = libcooling.get_lambda_from_Density_EnergyInt_FeH(rho, u, fe)

    dudt = l / rho
    dt = u / dudt
    print()
    print(("cooling time = ", dt))



###########################################################################################
#  package:   pNbody
#  file:      system.py
#  brief:     some useful system routines
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


def runCommand(cmd):
  """
  This function guarantee that the exit code will not be 0.
  Essential to really fail when os.system fails.
  """
  import os,sys
  
  exit_status = os.system(cmd)
  if exit_status != 0:
    print("runPythonScript: command\n\n%s\n\nfailed with exit status:"%cmd, exit_status) 
    sys.exit(sys.exit(os.waitstatus_to_exitcode(exit_status)))



#!/usr/bin/env python3
###########################################################################################
#  package:   Mockimgs
#  file:      __init__.py
#  brief:     init file
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from astropy import units as u

from . import instrument
from . import telescope
from . import filters
from . import ccd
from . import ifu
from . import mos
from . import obs_object



# list of telescopes      
telescopes = {}
telescopes["arrakihs"] = telescope.Telescope(name="iSIM-170",focal=1500*u.mm)
telescopes["euclid"]   = telescope.Telescope(name="Euclid"  ,focal=24.5*u.m)
telescopes["CTIO4.0m"]   = telescope.Telescope(name="CTIO4.0m"  ,focal=11.81*u.m)

# list of ccds
ccds = {}
ccds["arrakihs_vis"]   = ccd.CCD(name="arrakihs_vis",shape=[4096,4096],size=[41*u.mm,41*u.mm])
ccds["arrakihs_nir"]   = ccd.CCD(name="arrakihs_nir",shape=[2048,2048],size=[36.9*u.mm,36.9*u.mm])
ccds["CCD273-84"]      = ccd.CCD(name="CCD273-84"   ,shape=[4096,4096],pixel_size=[12*u.micron,12*u.micron])
ccds["CIS-300"]        = ccd.CCD(name="CIS-300"     ,shape=[4096,4096],pixel_size=[10*u.micron,10*u.micron])
ccds["Hawaii-2RG"]     = ccd.CCD(name="Hawaii-2RG"  ,shape=[2048,2048],pixel_size=[18*u.micron,18*u.micron])
ccds["DECam"]          = ccd.CCD(name="DECam"       ,shape=[2290,2290],pixel_size=[15*u.micron,15*u.micron])

#ccds["euclid_vis"]     = ccd.CCD(name="euclid_vis"  ,shape=[24576,24576])






# list of instruments
instruments = {}

#########################
# ARRAKIHS
#########################

###########
# Visible

# GAEA

instruments["arrakihs_vis_GAIA_VIS"] = instrument.Instrument(
  name        = "arrakihs_vis_GAIA_VIS",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("GAEA_VIS"),
  )

# BastI

instruments["arrakihs_vis_G"] = instrument.Instrument(
  name        = "arrakihs_vis_G",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_GAIA_G"),
  )

instruments["arrakihs_vis_G_RP"] = instrument.Instrument(
  name        = "arrakihs_vis_G_RP",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_GAIA_G_RP"),
  )
  
instruments["arrakihs_vis_G_BP"] = instrument.Instrument(
  name        = "arrakihs_vis_G_BP",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_GAIA_G_BP"),
  )  

instruments["arrakihs_vis_SDSSu"] = instrument.Instrument(
  name        = "arrakihs_vis_SDSSu",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_SDSS_u"),
  )
  
instruments["arrakihs_vis_SDSSg"] = instrument.Instrument(
  name        = "arrakihs_vis_SDSSg",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_SDSS_g"),
  )
  
instruments["arrakihs_vis_SDSSr"] = instrument.Instrument(
  name        = "arrakihs_vis_SDSSr",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_SDSS_r"),
  )

instruments["arrakihs_vis_SDSSi"] = instrument.Instrument(
  name        = "arrakihs_vis_SDSSi",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_SDSS_i"),
  )

instruments["arrakihs_vis_SDSSz"] = instrument.Instrument(
  name        = "arrakihs_vis_SDSSz",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_SDSS_z"),
  )
  
instruments["arrakihs_vis_F475X"] = instrument.Instrument(
  name        = "arrakihs_vis_F475X",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_HST_F475X"),
  )  
    
instruments["arrakihs_vis_VIS"] = instrument.Instrument(
  name        = "arrakihs_vis_VIS",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BastI_Euclid_VIS"),
  )    

# SB99
#instruments["arrakihs_vis_VIS1"] = instrument.Instrument(
#  name        = "arrakihs_vis_VIS1",
#  telescope   = telescopes["arrakihs"],
#  ccd         = ccds["CIS-300"],
#  filter_type = filters.Filter("SB99_ARK_VIS1"),
#  )

#instruments["arrakihs_vis_VIS2"] = instrument.Instrument(
#  name        = "arrakihs_vis_VIS2",
#  telescope   = telescopes["arrakihs"],
#  ccd         = ccds["CIS-300"],
#  filter_type = filters.Filter("SB99_ARK_VIS2"),
#  )

# BPASS
instruments["arrakihs_vis_VIS1"] = instrument.Instrument(
  name        = "arrakihs_vis_VIS1",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BPASS230_ARK_VIS1"),
  )

instruments["arrakihs_vis_VIS2"] = instrument.Instrument(
  name        = "arrakihs_vis_VIS2",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["CIS-300"],
  filter_type = filters.Filter("BPASS230_ARK_VIS2"),
  )






##################
# Near Infrared

# GAEA

instruments["arrakihs_nir_GAIA_Y"] = instrument.Instrument(
  name        = "arrakihs_nir_GAIA_Y",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["Hawaii-2RG"],
  filter_type = filters.Filter("GAEA_Y"),
  )

instruments["arrakihs_nir_GAIA_J"] = instrument.Instrument(
  name        = "arrakihs_nir_GAIA_J",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["Hawaii-2RG"],
  filter_type = filters.Filter("GAEA_J"),
  )

# BastI

instruments["arrakihs_nir_J"] = instrument.Instrument(
  name        = "arrakihs_nir_J",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["Hawaii-2RG"],
  filter_type = filters.Filter("BastI_Euclid_J"),
  )      

instruments["arrakihs_nir_Y"] = instrument.Instrument(
  name        = "arrakihs_nir_Y",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["Hawaii-2RG"],
  filter_type = filters.Filter("BastI_Euclid_Y"),
  )      

instruments["arrakihs_nir_H"] = instrument.Instrument(
  name        = "arrakihs_nir_H",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["Hawaii-2RG"],
  filter_type = filters.Filter("BastI_Euclid_H"),
  )   

# SB99

#instruments["arrakihs_nir_NIR1"] = instrument.Instrument(
#  name        = "arrakihs_nir_NIR1",
#  telescope   = telescopes["arrakihs"],
#  ccd         = ccds["Hawaii-2RG"],
#  filter_type = filters.Filter("SB99_ARK_NIR1"),
#  )   

#instruments["arrakihs_nir_NIR2"] = instrument.Instrument(
#  name        = "arrakihs_nir_NIR2",
#  telescope   = telescopes["arrakihs"],
#  ccd         = ccds["Hawaii-2RG"],
#  filter_type = filters.Filter("SB99_ARK_NIR2"),
#  )   
  
# BPASS (now default)

instruments["arrakihs_nir_NIR1"] = instrument.Instrument(
  name        = "arrakihs_nir_NIR1",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["Hawaii-2RG"],
  filter_type = filters.Filter("BPASS230_ARK_NIR1"),
  )   

instruments["arrakihs_nir_NIR2"] = instrument.Instrument(
  name        = "arrakihs_nir_NIR2",
  telescope   = telescopes["arrakihs"],
  ccd         = ccds["Hawaii-2RG"],
  filter_type = filters.Filter("BPASS230_ARK_NIR2"),
  )     
  
  
  

#########################
# DES DECam (SDSS filters)
#########################

instruments["DES_DECam_SDSSg"] = instrument.Instrument(
  name        = "DES_DECam_SDSSg",
  telescope   = telescopes["CTIO4.0m"],
  ccd         = ccds["DECam"],
  filter_type = filters.Filter("BastI_SDSS_g"),
  )

instruments["DES_DECam_SDSSr"] = instrument.Instrument(
  name        = "DES_DECam_SDSSr",
  telescope   = telescopes["CTIO4.0m"],
  ccd         = ccds["DECam"],
  filter_type = filters.Filter("BastI_SDSS_r"),
  )

instruments["DES_DECam_SDSSz"] = instrument.Instrument(
  name        = "DES_DECam_SDSSz",
  telescope   = telescopes["CTIO4.0m"],
  ccd         = ccds["DECam"],
  filter_type = filters.Filter("BastI_SDSS_z"),
  )

instruments["DES_DECam_SDSSi"] = instrument.Instrument(
  name        = "DES_DECam_SDSSi",
  telescope   = telescopes["CTIO4.0m"],
  ccd         = ccds["DECam"],
  filter_type = filters.Filter("BastI_SDSS_i"),
  )




def InstrumentsList():
  
  for name in instruments.keys():
    print(60*"#")
    print(name)
    print(60*"#")
    instruments[name].info()
    print()
    print()
  
  



''' 
 @package   pNbody
 @file      liblog.py
 @brief     Log File
 @copyright GPLv3
 @author    Yves Revaz <yves.revaz@epfl.ch>
 @section   COPYRIGHT  Copyright (C) 2017 EPFL (Ecole Polytechnique Federale de Lausanne)  LASTRO - Laboratory of Astrophysics of EPFL

 This file is part of pNbody. 
'''

import glob
import os
import sys
import string
import time
import traceback


try:
  import Tkinter as tk
  is_tk = True
except ImportError:
  is_tk = False 
  
  
###########################

class Log:

###########################

  '''
  a log class
  '''


  def __init__(self,directory,show='yes',append='no',filename=None,logframe=None):
    '''
    open the log file
    '''
    
    self.logframe = None
    self.filename = filename
        
    if show == 'yes':
      self.show = 1
    else:
      self.show = 0
      
    if not os.path.exists(directory):
      os.makedirs(directory)      
    
    t = time.localtime()
    if self.filename is None:
      #self.filename = "%00004d-%002d-%002dT%002d:%002d:%002d.log"%(t[0],t[1],t[2],t[3],t[4],t[5])
      self.filename = "log.dat"
    
    self.filename = os.path.join(directory,self.filename)
    
    if append == 'yes':
      self.f = open(self.filename,'a')
    else:
      self.f = open(self.filename,'w')   


  def write(self,line,name=None):
    '''
    write a line
    '''
        
    if self.show:
        
      if self.logframe is None:
        print(line)
      else:
        line = line.rstrip()
        line = line+'\n'

    if is_tk:
      self.logframe.config(state=tk.NORMAL)
      self.logframe.insert(tk.INSERT,line)
      self.logframe.yview(tk.MOVETO,1) 
      self.logframe.config(state=tk.DISABLED)
    else:
      print(line)  


    '''
    t = time.localtime()
    tnow = "%00004d-%002d-%002dT%002d:%002d:%002d"%(t[0],t[1],t[2],t[3],t[4],t[5])

    if name is not None:
       self.f.write("%s %s %s\n"%(tnow,name,line))     
    else:  
      self.f.write("%s %s\n"%(tnow,line)) 
    
    self.f.flush()
    '''
 
  def write_traceback(self):
    
    traceback.print_tb(sys.exc_info()[2],file=self.f)
    self.write_content()
      
  def write_content(self):
  
    if self.show:
      self.f.close()

      self.f = open(self.filename,'r')
      lines = self.f.readlines()
      self.f.close()
     
      for line in lines:
        self.write(line) 
	
      self.f = open(self.filename,'w')	
    

  def clear(self):
    if self.show:
      if self.logframe is not None:
        
        if is_tk:
          self.logframe.config(state=tk.NORMAL)
          self.logframe.delete(1.0,tk.END)
          #self.logframe.yview(MOVETO,1) 
          self.logframe.config(state=tk.DISABLED)
        else:
          pass
    
  def close(self):
    '''
    close the file
    '''
    
    self.f.close()
    
    
    



###########################################################################################
#  package:   pNbody
#  file:      auriga.py
#  brief:     AURIGA file format
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>, Loic Hausammann <loic_hausammann@hotmail.com>
#
# This file is part of pNbody.
###########################################################################################


##########################################################################
#
# GEAR HDF5 CLASS
#
##########################################################################

import numpy as np
import h5py

import pNbody
from pNbody import h5utils
from pNbody import mpi, error, units, h5utils

try:				# all this is useful to read files
    from mpi4py import MPI
except BaseException:
    MPI = None



class Nbody_gh5:
  
  
    def get_default_arrays_props(self):
      '''
      get the default properties of arrays considered
      This function is basically used to initialize self.arrays_props
      
      "h5name"  :     name in the hdf5 file
      "dtype"   :     numpy dtype                       
      "dim"     :     dimention 
      "ptypes"  :     type of particles that store this property (useless for the moment)
      "read"    :     read it by default or not
      "write"   :     write it by default or not
      "default" :     default values of the components
      "loaded"  :     is the array currently loaded    
      
      # position must always be first, this is a convention
      '''
      
      all_ptypes = list(range(self.get_mxntpe()))
      
      aprops = {}
      
      aprops["pos"] =  {
                        "h5name"  :     "Coordinates", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     3, 
                        "ptypes"  :     all_ptypes,
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["vel"] =  {
                        "h5name"  :     "Velocities", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     3,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                      
      aprops["mass"] =  {
                        "h5name"  :     "Masses", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     1,
                        "loaded"  :     False
                      }

      aprops["num"] =  {
                        "h5name"  :     "ParticleIDs", 
                        "dtype"   :     np.uint32,                        
                        "dim"     :     1,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                          


      aprops["u"] =  {
                        "h5name"  :     "InternalEnergy", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["rho"] =  {
                        "h5name"  :     "Density", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                      
      aprops["metallicity"] =  {
                        "h5name"  :     "GFM_Metallicity", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                      

      return aprops  
    
    

    def load(self,name=None,ptypes=None,force=False):
      '''
      The function in charge of loading arrays on demand.
      Here we use the generic hdf5 one.
      '''
      self.loadFromHDF5(name=name,ptypes=ptypes,force=force)

    
    def dump(self,aname=None):
      '''
      The function in charge of duming the array aname 
      to a file on demand.
      '''
      self.dumpToHDF5(aname=aname)
      
 

    def _init_spec(self):
        # create an empty header (in case the user is creating a dataset)
        self._header = []

    def get_excluded_extension(self):
        """ 
        Return a list of file to avoid when extending the default class.
        """
        return []

    def getParticleMatchingDict(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        
        index = {
            'gas':    0,
            'halo':   1,
            'disk':   2,
            'sink':   3,
            'stars':  4,
            'bndry':  2,
            'stars1': 4,
            'halo1':  1}
        
        return index 

    def check_spec_ftype(self):
        try:

            fd = h5utils.openFile(self.p_name_global[0]) 
            
 
            test =         "Header"     not in fd
            test = test or "Composition_vector_length" not in fd["Header"].attrs
            
            fd.close()
            
            if test:
                raise error.FormatError("auriga")

        except IOError as e:
            if self.verbose > 1:
                print("arepo not recognized: %s" % e)
            raise error.FormatError("auriga")

    def set_pio(self, pio):
        """ Overwrite function
        """
        pNbody.Nbody.set_pio(self, "no")

    def get_read_fcts(self):
        return [self.read_particles]

    def get_write_fcts(self):
        return [self.write_particles]

    def get_mxntpe(self):
        return 6

    def get_header_translation(self):
        """
        Gives a dictionnary containing all the header translation.
        If a new variable is possible in the HDF5 format, only the translation
        is required for the reader/writer.
        As h5py is not supporting dictionnary, they need special care when reading/writing.
        """
        # dict containing all the main header variables (=> easy acces)
        # e.g. self.npart will contain NumPart_ThisFile
        header_var = {}

        # Size variables
        header_var["Header/NumPart_ThisFile"] = "npart"
        header_var["Header/NumPart_Total"] =  "npart_tot"
        header_var["Header/NumPart_Total_HighWord"] = "nallhw"
        header_var["Header/MassTable"] = "massarr"
        header_var["Header/NumFilesPerSnapshot"] = "num_files"
        header_var["Header/BoxSize"] = "boxsize"
        header_var["Header/Flag_Entropy_ICs"] = "flag_entr_ics"
        header_var["Header/Composition_vector_length"] = "Composition_vector_length"

        # Physics
        header_var["Header/Scale-factor"] = "atime"
        header_var["Header/Redshift"] = "redshift"
        header_var["Header/Time"] = "time"

        # Cosmology
        #header_var["Cosmology/Omega_b"] = "omegab"
        header_var["Cosmology/Omega_m"] = "omega0"
        header_var["Cosmology/Omega_lambda"] = "omegalambda"
        header_var["Cosmology/h"] = "hubbleparam"
        header_var["Cosmology/Cosmological run"] = "cosmorun"
        


        # Units
        #header_var["Units/Unit velocity in cgs (U_V)"] = "UnitVelocity_in_cm_per_s"
        #header_var["Units/Unit length in cgs (U_L)"] = "UnitLength_in_cm"
        #header_var["Units/Unit mass in cgs (U_M)"] = "UnitMass_in_g"
        #header_var["Units/Unit time in cgs (U_t)"] = "Unit_time_in_cgs"
        #header_var["Units/Unit temperature in cgs (U_T)"] = "Unit_temp_in_cgs"
        #header_var["Units/Unit current in cgs (U_I)"] = "Unit_current_in_cgs"

        # Code
        header_var["Code/Code"] = "Code"
        header_var["Code/CFLAGS"] = "cflags"
        header_var["Code/Code Version"] = "code_version"
        header_var["Code/Compiler Name"] = "compiler_name"
        header_var["Code/Compiler Version"] = "compiler_version"
        header_var["Code/Configuration options"] = "config_options"
        header_var["Code/FFTW library version"] = "fftw_lib_version"
        header_var["Code/Git Branch"] = "git_branch"
        header_var["Code/Git Date"] = "git_date"
        header_var["Code/Git Revision"] = "git_revision"
        header_var["Code/HDF5 library version"] = "hdf5_lib_version"
        header_var["Code/MPI library"] = "mpi_lib"

        # HydroScheme
        header_var["HydroScheme/Adiabatic index"] = "adiabatic_index"
        header_var["HydroScheme/CFL parameter"] = "cfl_parameter"
        header_var["HydroScheme/Dimension"] = "dimension"
        header_var["HydroScheme/Kernel delta N_ngb"] = "kernel_delta_n_ngb"
        header_var["HydroScheme/Kernel eta"] = "kernel_eta"
        header_var["HydroScheme/Kernel function"] = "kernel_function"
        header_var["HydroScheme/Kernel target N_ngb"] = "kernel_target_n_ngb"
        header_var["HydroScheme/Max ghost iterations"] = "max_ghost_iterations"
        header_var["HydroScheme/Maximal smoothing length"] = "maximal_smoothing_length"
        header_var["HydroScheme/Scheme"] = "scheme"
        header_var["HydroScheme/Smoothing length tolerance"] = "smoothing_length_tolerance"
        header_var["HydroScheme/Thermal Conductivity Model"] = "thermal_conductivity_model"
        header_var["HydroScheme/Viscosity Model"] = "viscosity_model"
        header_var["HydroScheme/Viscosity alpha"] = "viscosity_alpha"
        header_var["HydroScheme/Viscosity beta"] = "viscosity_beta"
        header_var["HydroScheme/Volume log(max(delta h))"] = "volume_log"
        header_var["HydroScheme/Volume max change time-step"] = "volume_max_change"

        # Parameters
        #header_var["Parameters/InitialConditions:shift"] = "InitialConditions_shift"
        header_var["Header/Shift"] = "InitialConditions_shift"



        # Swift directory
        header_var["RuntimePars/PeriodicBoundariesOn"] = "periodic"
        
        
        # chemistry (this is now done elsewhere)
        #header_var["Parameters/GEARFeedback:elements"] = "ChimieElements"
        #header_var["SubgridScheme/Chemistry element count"] = "ChimieNelements"
        
        

        return header_var

    def get_list_header(self):
        """
        Gives a list of header directory from self.get_header_translation
        """
        list_header = []
        trans = self.get_header_translation()
        for key, tmp in list(trans.items()):
            directory = key.split("/")[0]
            if directory not in list_header:
                list_header.append(directory)
        return list_header



    def get_default_spec_vars(self):
        """
        return specific variables default values for the class
        """
        

        return {'massarr': np.array([0, 0, 0, 0, 0, 0]),
                'atime': 0.,
                'redshift': 0.,
                'flag_sfr': 0,
                'flag_feedback': 0,
                'npart_tot': np.array([0, 0, self.nbody, 0, 0, 0]),
                'npart': np.array([0, 0, self.nbody, 0, 0, 0]),
                'flag_cooling': 0,
                'num_files': 1,
                'boxsize': 0.,
                'omega0': 0.,
                'omegalambda': 0.,
                'hubbleparam': 0.,
                'flag_age': 0.,
                'flag_metals': 0.,
                'nallhw': np.array([0, 0, 0, 0, 0, 0]),
                'flag_entr_ics': 0,
                'flag_chimie_extraheader': 0,
                'critical_energy_spec': 0.,
                'empty': 48 * '',
                'comovingintegration': True,
                'hubblefactorcorrection': False,
                'comovingtoproperconversion': True,
                'ChimieNelements': 0,
                'utype':"auriga",
                "Composition_vector_length" : 0
                }


                  

    def get_massarr_and_nzero(self):
        """
        return massarr and nzero

        !!! when used in //, if only a proc has a star particle,
        !!! nzero is set to 1 for all cpu, while massarr has a length of zero !!!
        """

        if self.has_var('massarr') and self.has_var('nzero'):
          if self.massarr is not None and self.nzero is not None:
            self.message("warning : get_massarr_and_nzero : here we use massarr and nzero",self.massarr,self.nzero,verbosity=2)
            return self.massarr, self.nzero

        massarr = np.zeros(len(self.npart), float)
        nzero = 0

        # for each particle type, see if masses are equal
        for i in range(len(self.npart)):
            first_elt = sum((np.arange(len(self.npart)) < i) * self.npart)
            last_elt = first_elt + self.npart[i]

            if first_elt != last_elt:
                c = (self.mass[first_elt] ==
                     self.mass[first_elt:last_elt]).astype(int)
                if sum(c) == len(c):
                    massarr[i] = self.mass[first_elt]
                else:
                    nzero = nzero + len(c)

        return massarr.tolist(), nzero




    def read_particles(self, f):
        """
        read gadget file
        """
                
        from copy import deepcopy
        import time
                
        # go to the end of the file
        if f is not None:
            f.seek(0, 2)


        fd = h5utils.openFile(self.p_name_global[0],'r')    
            

        ################
        # read header
        ################
        self.message("reading header...")

        # set default values
        default = self.get_default_spec_vars()
        for key, i in list(default.items()):
            setattr(self, key, i)

        # get values from snapshot
        trans = self.get_header_translation()

        list_header = self.get_list_header()
        

        for name in list_header:
            if name not in fd:
                continue
                
            # e.g. create self.npart with the value
            # fd["Header"].attrs["NumPart_ThisFile"]
                        
            for key in fd[name].attrs:
                                                                    
              
                full_name = name + "/" + key
                if full_name not in list(trans.keys()):
                    trans[full_name] = full_name

                tmp = fd[name].attrs[key]
                if isinstance(tmp, bytes) and tmp == "None":
                    tmp = None
                if isinstance(tmp, bytes):
                    tmp = tmp.decode('utf-8')
                setattr(self, trans[full_name], tmp)
                
                self.message("%s %s >> %s %s"%(name,key,trans[full_name],tmp),verbosity=2)
        

        ################
        # read cosmorun
        ################        
        if self.has_var('cosmorun'):
          
          self.message("reading cosmorun...")
                    
          if self.cosmorun[0]==0:
            self.atime = self.time[0] 
            self.setComovingIntegrationOff()  # obsolete
            self.HubbleFactorCorrectionOff()
            self.ComovingToProperConversionOff()
                        
        
        
        ################
        # read SubgridScheme
        ################
        

        # get the chemical elements and solar abundances
        if "SubgridScheme" in fd:
          
          self.message("reading SubgridScheme...")
        
          subgridGrp = fd["SubgridScheme"]
          
          attrs = dict(subgridGrp.attrs)
          
          if "Chemistry element count" in attrs: 
            # this case is useful if SolarAbundances is not present
            self.ChimieNelements = int(attrs["Chemistry element count"][0])
            self._AM.arrays_props["metals"]["dim"] = self.ChimieNelements 
                        

          if "SolarAbundances" in subgridGrp: 

            d = dict(subgridGrp["SolarAbundances"].attrs)
        
            self.ChimieSolarMassAbundances = {}
        
            for key in d.keys():
              self.ChimieSolarMassAbundances[key] = d[key][0]
          
          
          if "NamedColumns" in subgridGrp:
                      
            tmp = subgridGrp["NamedColumns"]
            
            if "MetalMassFractions" in tmp:
              self.ChimieElements  = list(tmp["MetalMassFractions"][:])    
        
              for i,elt in enumerate(self.ChimieElements):
                self.ChimieElements[i] = self.ChimieElements[i].decode("utf-8")
        
              self.ChimieNelements = len(self.ChimieElements)
            
              # we update 
              self._AM.arrays_props["metals"]["dim"] = self.ChimieNelements 
            
          

        
        ################
        # read particles
        ################
        
        self.message("reading particles...")
                
        # get npart from the blocks
        self.npart = h5utils.get_npart_from_dataset(self.p_name_global[0],self._AM.array_h5_key('pos'),self.get_mxntpe(),ptypes=self.ptypes)
                
        # check that arrays are present
        self._AM.setReadableArraysFromHDFH5File(self.p_name_global[0],arrays=self.arrays,ptypes=self.ptypes)
        
        # load all arrays according to the information
        # provided by self.arrays_props
        self.read_arrays(ptypes=self.ptypes)

        # set tpe
        self.tpe = np.array([], np.int32)
        for i in range(len(self.npart)):
            self.tpe = np.concatenate((self.tpe, np.ones(self.npart[i]) * i))

        # compute nzero
        nzero = 0
        mass = np.array([])

        for i in range(len(self.npart)):
            if self.massarr[i] == 0:
                nzero = nzero + self.npart[i]
            else:
                self.warning("Massarr is not supported! Please specify the mass of all the particles!",verbosity=2)

        self.nzero = nzero


        #############################
        # specific final conversions
        #############################        
        
        
        if hasattr(self, 'idp'):
          self.idp = self.idp.astype(int) 
        
        if type(self.atime) == np.ndarray:
          self.atime = self.atime[0]
          
        if type(self.redshift) == np.ndarray:
          self.redshift = self.redshift[0]
        
        if type(self.hubbleparam) == np.ndarray:
          self.hubbleparam = self.hubbleparam[0]
          
        if type(self.omega0) == np.ndarray:
          self.omega0 = self.omega0[0]        

        if type(self.omegalambda) == np.ndarray:
          self.omegalambda = self.omegalambda[0]   
          
          
          
                    

    def write_particles(self, f):
        """
        specific format for particle file
        """
        # go to the end of the file
        if f is not None:
            f.seek(0, 2)
            

        name = "Unit_temp_in_cgs"
        if not hasattr(self, name):
            setattr(self, name, 1.0)

        name = "Unit_current_in_cgs"
        if not hasattr(self, name):
            setattr(self, name, 1.0)

        import h5py
        # not clean, but work around pNbody
        filename = self.p_name_global[0]
        # open file
        h5f = h5utils.openFile(filename,'w')    

        # add units to the usual gh5 struct
        if hasattr(self, "unitsparameters"):
            units = self.unitsparameters.get_dic()
            for key, i in list(units.items()):
                if not hasattr(self, key):
                    setattr(self, key, i)

        if hasattr(self,"UnitVelocity_in_cm_per_s") and hasattr(self,"UnitLength_in_cm"):
          self.Unit_time_in_cgs = self.UnitLength_in_cm / self.UnitVelocity_in_cm_per_s
        
 
          

        ############
        # HEADER
        ############
        self.message("Writing header...")

        list_header = self.get_list_header()
        trans = self.get_header_translation()
        
        for name in list_header:
            h5f.create_group(name)
                    
            
        for key in self.get_list_of_vars():
          
            if key in list(trans.values()):
                            
                ind = list(trans.values()).index(key)
                name, hdf5 = list(trans.keys())[ind].split("/")
                value = getattr(self, key)
                
                # !!! force value of npart to be npart_tot (to be improved)
                if mpi.mpi_NTask() > 1: 
                  if key=="npart":
                    value = self.npart_tot
            
                if type(value) is not str:
                  value = np.array(value)
                else:
                  value = np.array(value,dtype="S")  
                  
                if value.shape == ():
                  value = np.array([value])
                  
                
                if not isinstance(value, dict):
                    if value is None:
                        h5f[name].attrs[hdf5] = "None"
                    else:
                        h5f[name].attrs[hdf5] = value
                        

        

        ##############
        # SubgridScheme
        ##############
        self.message("Writing SubgridScheme...")
        
        if self.has_var("ChimieSolarMassAbundances"):
                
          subgridGrp = h5f.create_group("SubgridScheme")
          tmp = h5f.create_group("SubgridScheme/SolarAbundances")
        
          for key, val in self.ChimieSolarMassAbundances.items():
            tmp.attrs[key] = np.array([val],np.float32)

          tmp = h5f.create_group("SubgridScheme/NamedColumns")
          asciiList = [n.encode("ascii", "ignore") for n in self.ChimieElements]
          tmp.create_dataset('MetalMassFractions', (len(asciiList),),'S10', asciiList)            
            
            
                 


        h5utils.closeFile(h5f)
        
        ##############
        # PARTICULES
        ##############
        self.message("Writing particles...")
            
            
        # loop over all arrays
        for aname in self.get_list_of_arrays():
          if aname in self._AM.arrays():
          
            if self._AM.array_write(aname):
              # dump the array
              self.dump(aname)
            else:
              self.warning("array %s is not set as to be written."%aname,verbosity=2)
          
          else:
            self.warning("array %s is not stored in the array manager."%aname,verbosity=2)    
              



#!/usr/bin/python

import numpy as np
from matplotlib import pyplot as plt
import sys as sys

try:
  import scipy.integrate as integ
  import scipy.optimize as opt
  is_scipy = True
except ImportError:
  is_scipy = False

class Sedov:
  """
  This class is a routine to compute a sedov solution following the implementation of Kamm 2000
  """
  def __init__(self,E_0=10., rho_0=0.1,gamma=5./3., j=3., omega=0.,Np=100):
    
    print("initialisation of the Sedov computation")
    ###########################################
    # init the constants for the computations
    ###########################################

    self.E_0 	= E_0 	#injected energy
    self.rho_0 	= rho_0	#medium density
    self.gamma	= gamma	#adiabatic index
    self.j 		= j	#simulation dimension (1,2 or 3)
    self.omega 	= omega #power for the density profile \rho=\rho_0 r^{-\omega}
    self.Np		= Np 	#number of point for the spatial computation
    
    ##################################
    # compute other useful parameters
    ##################################
    
    
    self.a = (self.j + 2. - self.omega) * (self.gamma + 1.) / 4.
    
    self.b = (self.gamma + 1.) / (self.gamma - 1.)
    
    self.c = (self.j + 2. - self.omega) * self.gamma / 2.
    
    self.d = (self.j + 2. - self.omega) * (self.gamma + 1.) / (( self.j + 2. - self.omega) * (self.gamma + 1.) - 2. * (2. + self.j * (self.gamma - 1.)))
    
    self.e = (2. + self.j * (self.gamma - 1.))/ 2.
    
    
    self.alpha_0 = 2. / (self.j + 2. - self.omega)
    
    self.alpha_2 = - (self.gamma - 1.)/ ( 2.*(self.gamma-1.) + self.j - self.gamma * self.omega)
    
    self.alpha_1 = (self.j + 2. - self.omega) *self.gamma / (2. + self.j*(self.gamma-1.)) * ( 2. * (self.j*(2.-self.gamma) - self.omega)/ (self.gamma * (self.j + 2. - self.omega)**2.) - self.alpha_2)
    
    self.alpha_3 = (self.j - self.omega) / ( 2*(self.gamma-1) + self.j - self.gamma * self.omega)
    
    self.alpha_4 = (self.j + 2. - self.omega) * (self.j - self.omega)/ (self.j*(2.-self.gamma) - self.omega) * self.alpha_1
    
    self.alpha_5 = (self.omega * (1. + self.gamma) - 2.*self.j) / (self.j*(2.-self.gamma) - self.omega)
    
    #min and max velocity
    self.V_0 = 2. / (self.j + 2. - self.omega) / self.gamma
    self.V_2 = 4./ (j + 2. - self.omega) / (self.gamma + 1.)
    
    #compute the alpha coeff now
    self.compute_alpha()


  def J_1_int(self,V):
    # integrand of the J_1 energy integral
    return - (self.gamma + 1.) / (self.gamma - 1.) * V**2. * ( self.alpha_0 / V + self.alpha_2 * self.c / (self.c*V - 1. ) - self.alpha_1 * self.e / (1. - self.e * V) )* ( (self.a* V)**self.alpha_0 *(self.b * (self.c*V -1.))**self.alpha_2 * (self.d*(1. - self.e*V))**self.alpha_1)**(-(self.j+2.-self.omega)) * (self.b*(self.c*V - 1.))**self.alpha_3 * (self.d*(1. - self.e*V))**self.alpha_4 * (self.b*(1. - self.c/self.gamma*V))**self.alpha_5

  def J_2_int(self,V):
    # integrand of the J_1 energy integral
    return - (self.gamma + 1.) / (2. * self.gamma) * V**2. * (self.c* V - self.gamma)/ (1. - self.c*V) * ( self.alpha_0 / V + self.alpha_2 * self.c / (self.c*V - 1. ) - self.alpha_1 * self.e / (1. - self.e* V)) * ( (self.a* V)**self.alpha_0 *(self.b * (self.c*V -1.))**self.alpha_2 * (self.d*(1. - self.e*V))**self.alpha_1)**(-(self.j+2.-self.omega)) * (self.b*(self.c*V - 1.))**self.alpha_3 * (self.d*(1. - self.e*V))**self.alpha_4 * (self.b*(1. - self.c/self.gamma*V))**self.alpha_5

  def compute_alpha(self):
    #compute the alpha parameter in relation with energy conservation
    
    #compute the integrals form V_0 to V_2
    J_1 = integ.quad(self.J_1_int, self.V_0,self.V_2)
    J_2 = integ.quad(self.J_2_int, self.V_0,self.V_2)
    
    #compute the I's, WARNING FALSE for j=1 !!!!!! needs some delta
    I_1 = 2.**(self.j-2.) * ( np.pi * J_1[0])
    I_2 = 2.**(self.j-1.) / (self.gamma - 1.) * ( np.pi * J_2[0])
    
    self.alpha = I_1 + I_2

  def compute_rankin(self,t):
    #compute the values at the shock position with the rankin-hugenot boundary conditions
    self.r_2 	= (self.E_0 / (self.alpha * self.rho_0))**(1./(self.j+2.- self.omega)) * t**(2./(self.j + 2.- self.omega)) #radius of the shock front
    self.U 		= (2./ ( self.j + 2. - self.omega)) * self.r_2/t #velocity of the shock front
    self.v_2 	= (2. / (self.gamma + 1.))* self.U #velocity of the post-shock medium (just after the shock)
    self.rho_1 	= self.rho_0 * self.r_2**(-self.omega) #density of the initial medium
    self.rho_2 	= ((self.gamma + 1.) / (self.gamma - 1.)) * self.rho_1 #density of the post-shock medium (just after the shock)
    self.p_2 	= (2. / (self.gamma + 1.)) * self.rho_1 * self.U**2. #pressure just after the shock

  #here is some other useful parameter combination for the spatial solution
  def x_1(self,V):
    return self.a * V

  def x_2(self,V):
    return self.b * ( self.c*V - 1.)

  def x_3(self,V):
    return self.d*(1. - self.e*V)

  def x_4(self,V):
    return self.b*(1.- self.c/self.gamma * V)

  #spatial solution as a function of V
  def r(self,V):
    #radius
    return self.r_2 * self.x_1(V)**(-self.alpha_0) * self.x_2(V)**(-self.alpha_2) * self.x_3(V)**(-self.alpha_1)

  def v(self,V):
    #velocity
    return self.v_2 * self.x_1(V) * self.r(V) / self.r_2

  def rho(self,V):
    #density
    return self.rho_2 * self.x_1(V)**(self.alpha_0*self.omega) * self.x_2(V)**(self.alpha_3 + self.alpha_2*self.omega) * self.x_3(V)**(self.alpha_4 + self.alpha_1*self.omega) * self.x_4(V)**self.alpha_5

  def p(self,V):
    #pressure
    return self.p_2 * self.x_1(V)**(self.alpha*self.j) * self.x_3(V)**(self.alpha_4 + self.alpha_1*(self.omega-2.)) * self.x_4(V)**(1.+self.alpha_5)


  def compute_V(self):
    # it compute the V array correponding to the positions self.R, as it is implicit it needs a root finding method
    
    def V_find(V,R):
      #function to minimize
      return abs(R-self.r(V))
    
    #compute the V array HOW TO AVOID THE LOOP!!!!!!
    self.V=np.empty(self.Np)
    for i in range(self.Np):
            #self.V[i] = opt.fmin(V_find,self.V_0,args=[self.R[i]],disp=0)
      self.V[i] = opt.fmin(V_find,self.V_0,args=(self.R[i],),disp=0)

  def set_time(self,t):
    self.t = t #set the time
    
    self.compute_rankin(t) #compute the boundary condition at that time
    self.R = np.linspace(0,self.r_2,self.Np) #create the radius array from 0 to the shock front 
    
    self.compute_V() #find the corresponding V array

  def get_v(self):
    return self.v(self.V)

  def get_rho(self):
    return self.rho(self.V)

  def get_p(self):
    return self.p(self.V)

  def get_r(self):
    return self.R


if __name__ == '__main__':
  sedov=Sedov()
  sedov.set_time(0.07)
  R=sedov.get_r()
  v=sedov.get_v()
  p=sedov.get_p()
  rho=sedov.get_rho()
  
  plt.figure(figsize=(5,4))
  plt.plot(R,v/sedov.v_2,label=r'\rm{v}_\rm{r}')
  plt.plot(R,rho/sedov.rho_2,label=r'$\rho')
  plt.plot(R,p/sedov.p_2,label=r'$\rm{P}')
  plt.ylabel(r'$\rm{Normalized\ quantities')
  plt.xlabel(r'$R/R_s$')
  plt.axis([0.0,1.0,0.0,1.0])
  plt.legend(loc="best")
  plt.show()



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      __init__.py
#  brief:     init file
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################



"""
This python module is useful to manipulate N-body data.
It allows to compute simple physical values like energy,
kinetic momentum, inertial momentum, centre of mass etc.
It also allows to modify the data with rotation, translations,
select some particles, add particles etc.
Associated scripts like "gdisp", "mkmovie" or "movie" allow
to visualise the N-body data in different ways : surface density,
velocity map, velocity dispersion map, etc.

Yves Revaz 14.05.05
"""

# nbody classes
from .main import *

from ._version import get_versions
__version__ = get_versions()['full-revisionid']



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      einasto.py
#  brief:     Defines einastro profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
Einasto model
"""

import numpy as np

def Potential(rho0, a, r, m=6.,G=1.):
    """
    Potential
    
    Parameters
    ----------
    rho0 : specific density
    a    : scale length
    r    : radius
    m    : power (default=6)
    G    : gravitational constant (default=1)
    """
    pass
    

def Density(rho0, a, r, m=6.,G=1.):
    """
    Density
    
    Parameters
    ----------
    rho0 : specific density
    a    : scale length
    r    : radius
    m    : power (default=6)
    G    : gravitational constant (default=1)
    """
    return rho0 * np.exp(-(2*m)*((r/a)**(1./m)-1))
    

def Vcirc(rho0, a, r, m=6., G=1.):
    """
    Circular velocity
    
    Parameters
    ----------
    rho0 : specific density
    a    : scale length
    r    : radius
    m    : power (default=6)
    G    : gravitational constant (default=1)
    """
    pass

